/*
 * Copyright (C) 2005-2006 by CERN
 */

/*
 * @(#)$RCSfile: srm_ifce.c,v $ $Revision: 1.51 $ $Date: 2009/02/25 13:38:08 $ CERN Jean-Philippe Baud
 */

#include <sys/types.h>
#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <strings.h>
#include <sys/stat.h>
#include "gfal_api.h"
#include "gfal_internals.h"
#include "srmH.h"
#include "ISRM.nsmap"
#ifdef GFAL_SECURE
#include "cgsi_plugin.h"
#endif
#include "srmC.c"
#include "srmClient.c"

static const char gfal_remote_type[] = "SE";

static int srm_set_xfer_status (const char *, const char *, int, int, char *, int, int);
static int errorstring2errno (const char *);


srm_deletesurls (int nbfiles, const char **surls, const char *srm_endpoint,
		struct srm_filestatus **filestatuses, char *errbuf, int errbufsz, int timeout)
{
	struct srm1__advisoryDeleteResponse out;
	int i, ret;
	int sav_errno = 0;
	int flags;
	struct soap soap;
	struct ArrayOfstring surlarray;
	const char srmfunc[] = "advisoryDelete";

	soap_init (&soap);

#ifdef GFAL_SECURE
	flags = CGSI_OPT_DISABLE_NAME_CHECK;
	soap_register_plugin_arg (&soap, client_cgsi_plugin, &flags);
#endif

	soap.send_timeout = gfal_get_timeout_sendreceive ();
	soap.recv_timeout = gfal_get_timeout_sendreceive ();
	soap.connect_timeout = gfal_get_timeout_connect ();


	/* issue "advisoryDelete" request */

	surlarray.__size = nbfiles;
	surlarray.__ptr = (char **) surls;

	if ((ret = soap_call_srm1__advisoryDelete (&soap, srm_endpoint,
				srmfunc, &surlarray, &out))) {
		if(soap.fault != NULL && soap.fault->faultstring != NULL) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: %s",
					gfal_remote_type, srmfunc, srm_endpoint, soap.fault->faultstring);
			sav_errno = errorstring2errno (soap.fault->faultstring);
		} else if (soap.error == SOAP_EOF) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Connection fails or timeout",
					gfal_remote_type, srmfunc, srm_endpoint);
			sav_errno = ECOMM;
		} else {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Unknown SOAP error (%d)",
					gfal_remote_type, srmfunc, srm_endpoint, soap.error);
			sav_errno = ECOMM;
		}
		soap_end (&soap);
		soap_done (&soap);
		errno = sav_errno;
		return (-1);
	}
	soap_end (&soap);
	soap_done (&soap);

	if ((*filestatuses = (struct srm_filestatus *) calloc (nbfiles, sizeof (struct srm_filestatus))) == NULL) {
		errno = ENOMEM;
		return (-1);
	}
	for (i = 0; i < nbfiles; ++i) {
		memset (*filestatuses + i, 0, sizeof (struct srm_filestatus));
		(*filestatuses)[i].surl = strdup (surls[i]);
	}
	return (nbfiles);
}

srm_get (int nbfiles, const char **surls, int nbprotocols, char **protocols,
		int *reqid, char **token, struct srm_filestatus **filestatuses, int timeout)
{
	return (srm_getx (nbfiles, surls, nbprotocols, protocols, reqid,
				filestatuses, NULL, 0, timeout));
}

srm_getx (int nbfiles, const char **surls, int nbprotocols, char **protocols,
		int *reqid, struct srm_filestatus **filestatuses,
		char *errbuf, int errbufsz, int timeout)
{
	int i = 0;
	char **se_types;
	char **se_endpoints;
	char *srm_endpoint = NULL;

	if (setypesandendpointsfromsurl (surls[0], &se_types, &se_endpoints, errbuf, errbufsz) < 0)
		return (-1);

	while (se_types[i]) {
		if ((strcmp (se_types[i], "srm_v1")) == 0) 
			srm_endpoint = se_endpoints[i];
		i++;
	}

	free (se_types);
	free (se_endpoints);

	if (! srm_endpoint) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
				"[GFAL][srm_getx][EINVAL] %s: No matching SRMv1-compliant SE", surls[0]);
		errno = EINVAL;
		return (-1);
	}

	return (srm_getxe (nbfiles, surls, srm_endpoint, protocols, reqid,
				filestatuses, errbuf, errbufsz, timeout));
}

srm_getxe (int nbfiles, const char **surls, const char *srm_endpoint,
		char **protocols, int *reqid, struct srm_filestatus **filestatuses,
		char *errbuf, int errbufsz, int timeout)
{
	int flags;
    struct srmDiskCache__RequestFileStatus *f;
	int i, n;
	int nbproto = 0;
	struct srm1__getResponse outg;
	struct ArrayOfstring protoarray;
	struct srmDiskCache__RequestStatus *reqstatp;
	int ret;
	int sav_errno = 0;
	struct soap soap;
	struct ArrayOfstring surlarray;
	const char srmfunc[] = "get";

	soap_init (&soap);

#ifdef GFAL_SECURE
	flags = CGSI_OPT_DISABLE_NAME_CHECK;
	soap_register_plugin_arg (&soap, client_cgsi_plugin, &flags);
#endif

	soap.send_timeout = gfal_get_timeout_sendreceive ();
	soap.recv_timeout = gfal_get_timeout_sendreceive ();
	soap.connect_timeout = gfal_get_timeout_connect ();
   
	if (!protocols) {
		protocols = get_sup_proto ();
		while (protocols[nbproto] && *protocols[nbproto]) nbproto++;
		/* gsiftp is not included in the supported protocol list, but has to be in the default
		 * list for 'get' operation
		 * Note that the list returned by get_sup_proto has one extra place for that! */
		protocols[nbproto++] = "gsiftp";
		protocols[nbproto] = "";
	} else {
		while (protocols[nbproto] && *protocols[nbproto]) nbproto++;
		if (!protocols[nbproto]) protocols[nbproto] = "";
	}

	surlarray.__ptr = (char **)surls;
	surlarray.__size = nbfiles;
	protoarray.__ptr = protocols;
	protoarray.__size = nbproto;

	if ((ret = soap_call_srm1__get (&soap, srm_endpoint, srmfunc, &surlarray,
				&protoarray, &outg))) {
		if(soap.fault != NULL && soap.fault->faultstring != NULL) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: %s",
					gfal_remote_type, srmfunc, srm_endpoint, soap.fault->faultstring);
			sav_errno = errorstring2errno (soap.fault->faultstring);
		} else if (soap.error == SOAP_EOF) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
					"[%s][%s][] %s: Connection fails or timeout", gfal_remote_type, srmfunc, srm_endpoint);
			sav_errno = ECOMM;
		} else {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
					"[%s][%s][] %s: Unknown SOAP error (%d)", gfal_remote_type, srmfunc, srm_endpoint, soap.error);
			sav_errno = ECOMM;
		}
		soap_end (&soap);
		soap_done (&soap);
		errno = sav_errno;
		return (-1);
	}

	if ((reqstatp = outg._Result) == NULL || reqstatp->fileStatuses == NULL) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
				"[%s][%s][] %s: <empty response>", gfal_remote_type, srmfunc, srm_endpoint);
		soap_end (&soap);
		soap_done (&soap);
		errno = ECOMM;
		return (-1);
	}
	*reqid = reqstatp->requestId;

	n = reqstatp->fileStatuses->__size;

	if ((*filestatuses = calloc (n, sizeof(struct srm_filestatus))) == NULL) {
		soap_end (&soap);
		soap_done (&soap);
		errno = ENOMEM;
		return (-1);
	}

	for (i = 0; i < n; i++) {
		f = reqstatp->fileStatuses->__ptr[i];
		if (f->SURL)
			(*filestatuses)[i].surl = strdup (f->SURL);
		if (f->state) {
			if (strcasecmp (f->state, "pending") == 0)
				(*filestatuses)[i].status = 0;
			else if (strcasecmp (f->state, "failed") == 0)
				(*filestatuses)[i].status = -1;
			else
				(*filestatuses)[i].status = 1;
		}
		(*filestatuses)[i].fileid = f->fileId;
		if (f->TURL)
			(*filestatuses)[i].turl = strdup (f->TURL);
	}
	soap_end (&soap);
	soap_done (&soap);
	return (n);
}

srm_getstatus (int nbfiles, const char **surls, int reqid, char *token,
		struct srm_filestatus **filestatuses, int timeout)
{
	return (srm_getstatusx (nbfiles, surls, reqid, filestatuses, NULL, 0, timeout));
}

srm_getstatusx (int nbfiles, const char **surls, int reqid, struct srm_filestatus **filestatuses,
		char *errbuf, int errbufsz, int timeout)
{
	int i = 0;
	char **se_types;
	char **se_endpoints;
	char *srm_endpoint = NULL;

	if (setypesandendpointsfromsurl (surls[0], &se_types, &se_endpoints, errbuf, errbufsz) < 0)
		return (-1);

	while (se_types[i]) {
		if ((strcmp (se_types[i], "srm_v1")) == 0) 
			srm_endpoint = se_endpoints[i];
		i++;
	}

	free (se_types);
	free (se_endpoints);

	if (!srm_endpoint) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[GFAL][srm_getstatusx][EINVAL] %s: No matching SRMv1-compliant SE", surls[0]);
		errno = EINVAL;
		return (-1);
	}

	return (srm_getstatusxe (reqid, srm_endpoint, filestatuses, errbuf, errbufsz, timeout));
}

srm_getstatusxe (int reqid, const char *srm_endpoint, struct srm_filestatus **filestatuses,
		char *errbuf, int errbufsz, int timeout)
{
	int flags;
	struct srmDiskCache__RequestFileStatus *f;
	int i;
	int n;
	struct srm1__getRequestStatusResponse outq;
    struct srmDiskCache__RequestStatus *reqstatp;
	int ret;
	int sav_errno = 0;
	struct soap soap;
	const char srmfunc[] = "getRequestStatus";

	soap_init (&soap);

#ifdef GFAL_SECURE
	flags = CGSI_OPT_DISABLE_NAME_CHECK;
	soap_register_plugin_arg (&soap, client_cgsi_plugin, &flags);
#endif

	soap.send_timeout = gfal_get_timeout_sendreceive ();
	soap.recv_timeout = gfal_get_timeout_sendreceive ();
	soap.connect_timeout = gfal_get_timeout_connect ();


	if ((ret = soap_call_srm1__getRequestStatus (&soap, srm_endpoint,
				srmfunc, reqid, &outq))) {
		if(soap.fault != NULL && soap.fault->faultstring != NULL) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: %s",
					gfal_remote_type, srmfunc, srm_endpoint, soap.fault->faultstring);
			sav_errno = errorstring2errno (soap.fault->faultstring);
		} else if (soap.error == SOAP_EOF) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Connection fails or timeout",
					gfal_remote_type, srmfunc, srm_endpoint);
			sav_errno = ECOMM;
		} else {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Unknown SOAP error (%d)",
					gfal_remote_type, srmfunc, srm_endpoint, soap.error);
			sav_errno = ECOMM;
		}
		soap_end (&soap);
		soap_done (&soap);
		errno = sav_errno;
		return (-1);
	}

	if ((reqstatp = outq._Result) == NULL || reqstatp->fileStatuses == NULL) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
				"[%s][%s][] %s: <empty response>", gfal_remote_type, srmfunc, srm_endpoint);
		soap_end (&soap);
		soap_done (&soap);
		errno = ECOMM;
		return (-1);
	}

	n = reqstatp->fileStatuses->__size;

	if ((*filestatuses = (struct srm_filestatus *) calloc (n, sizeof(struct srm_filestatus))) == NULL) {
		soap_end (&soap);
		soap_done (&soap);
		errno = ENOMEM;
		return (-1);
	}

	for (i = 0; i < n; i++) {
		f = reqstatp->fileStatuses->__ptr[i];
		if (f->SURL)
			(*filestatuses)[i].surl = strdup (f->SURL);
		if (f->state) {
			if (strcasecmp (f->state, "pending") == 0)
				(*filestatuses)[i].status = 0;
			else if (strcasecmp (f->state, "failed") == 0)
				(*filestatuses)[i].status = -1;
			else
				(*filestatuses)[i].status = 1;
		}
		(*filestatuses)[i].fileid = f->fileId;
		if (f->TURL)
			(*filestatuses)[i].turl = strdup (f->TURL);
	}
	soap_end (&soap);
	soap_done (&soap);
	return (n);
}

srm_turlsfromsurls (int nbfiles, const char **surls, const char *srm_endpoint, GFAL_LONG64 *filesizes, char **protocols,
		int oflag, int *reqid, struct srm_filestatus **statuses, char *errbuf, int errbufsz, int timeout)
{
	int flags;
	int i, n;
	int nbproto = 0;
	struct srm1__getResponse outg;
	struct srm1__putResponse outp;
	struct srm1__getRequestStatusResponse outq;
	struct ArrayOfboolean permarray;
	struct ArrayOfstring protoarray;
	struct srmDiskCache__RequestStatus *reqstatp;
	int ret;
	int sav_errno = 0;
	int sleep_time, nbretries;
	struct ArrayOflong sizearray;
	struct soap soap;
	struct ArrayOfstring srcarray;
	struct ArrayOfstring surlarray;
	time_t endtime = 0;
	const char srmfunc_get[] = "get";
	const char srmfunc_put[] = "put";
	const char srmfunc_status[] = "getRequestStatus";
	char *srmfunc;

	soap_init (&soap);

#ifdef GFAL_SECURE
	flags = CGSI_OPT_DISABLE_NAME_CHECK;
	soap_register_plugin_arg (&soap, client_cgsi_plugin, &flags);
#endif

	soap.send_timeout = gfal_get_timeout_sendreceive ();
	soap.recv_timeout = gfal_get_timeout_sendreceive ();
	soap.connect_timeout = gfal_get_timeout_connect ();
   
	if (!protocols)
		protocols = get_sup_proto ();

	while (protocols[nbproto] && *protocols[nbproto]) nbproto++;
	if (!protocols[nbproto]) protocols[nbproto] = "";

	/* issue "get" or the "put" request */

	surlarray.__ptr = (char **)surls;
	surlarray.__size = nbfiles;
	protoarray.__ptr = protocols;
	protoarray.__size = nbproto;

	if ((oflag & O_ACCMODE) == 0) {
		srmfunc = (char *) srmfunc_get;

		if ((ret = soap_call_srm1__get (&soap, srm_endpoint, srmfunc, &surlarray,
					&protoarray, &outg))) {
			if(soap.fault != NULL && soap.fault->faultstring != NULL) {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: %s",
						gfal_remote_type, srmfunc, srm_endpoint, soap.fault->faultstring);
				sav_errno = errorstring2errno (soap.fault->faultstring);
			} else if (soap.error == SOAP_EOF) {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Connection fails or timeout",
						gfal_remote_type, srmfunc, srm_endpoint);
				sav_errno = ECOMM;
			} else {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Unknown SOAP error (%d)",
						gfal_remote_type, srmfunc, srm_endpoint, soap.error);
				sav_errno = ECOMM;
			}
			soap_end (&soap);
			soap_done (&soap);
			errno = sav_errno;
			return (-1);
		}
		reqstatp = outg._Result;
	} else {
		srmfunc = (char *) srmfunc_put;
		srcarray.__ptr = (char **)surls;
		srcarray.__size = nbfiles;
		sizearray.__ptr = filesizes;
		sizearray.__size = nbfiles;
		if ((permarray.__ptr =
					soap_malloc (&soap, nbfiles * sizeof(enum xsd__boolean))) == NULL) {
			soap_end (&soap);
			soap_done (&soap);
			errno = ENOMEM;
			return (-1);
		}
		for (i = 0; i< nbfiles; i++)
			permarray.__ptr[i] = true_;
		permarray.__size = nbfiles;
		if ((ret = soap_call_srm1__put (&soap, srm_endpoint, srmfunc, &srcarray,
					&surlarray, &sizearray, &permarray, &protoarray, &outp))) {
			if(soap.fault != NULL && soap.fault->faultstring != NULL) {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: %s",
						gfal_remote_type, srmfunc, srm_endpoint, soap.fault->faultstring);
				sav_errno = errorstring2errno (soap.fault->faultstring);
			} else if (soap.error == SOAP_EOF) {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Connection fails or timeout",
						gfal_remote_type, srmfunc, srm_endpoint);
				sav_errno = ECOMM;
			} else {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Unknown SOAP error (%d)",
						gfal_remote_type, srmfunc, srm_endpoint, soap.error);
				sav_errno = ECOMM;
			}
			soap_end (&soap);
			soap_done (&soap);
			errno = sav_errno;
			return (-1);
		}
		reqstatp = outp._Result;
	}
	if (reqstatp == NULL || reqstatp->fileStatuses == NULL) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: <empty response>",
				gfal_remote_type, srmfunc, srm_endpoint);
		soap_end (&soap);
		soap_done (&soap);
		errno = ECOMM;
		return (-1);
	}
	*reqid = reqstatp->requestId;

	/* wait for file "ready" */

	if (timeout > 0)
		endtime = time(NULL) + timeout;

	sleep_time = 5;
	nbretries = 0;

	while (reqstatp && strcasecmp (reqstatp->state, "pending") == 0) {
		if (reqstatp->retryDeltaTime > 0)
			sleep (reqstatp->retryDeltaTime);
		else {
			sleep (sleep_time);
			if (++nbretries < 8)
				sleep_time *= 2; // exponentially increase sleeping time, until ~10min
		}
		if ((ret = soap_call_srm1__getRequestStatus (&soap, srm_endpoint,
					srmfunc_status, *reqid, &outq))) {
			if(soap.fault != NULL && soap.fault->faultstring != NULL) {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: %s",
						gfal_remote_type, srmfunc_status, srm_endpoint, soap.fault->faultstring);
				sav_errno = errorstring2errno (soap.fault->faultstring);
			} else if (soap.error == SOAP_EOF) {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Connection fails or timeout",
						gfal_remote_type, srmfunc_status, srm_endpoint);
				sav_errno = ECOMM;
			} else {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Unknown SOAP error (%d)",
						gfal_remote_type, srmfunc_status, srm_endpoint, soap.error);
				sav_errno = ECOMM;
			}
			soap_end (&soap);
			soap_done (&soap);
			errno = sav_errno;
			return (-1);
		}
		reqstatp = outq._Result;

		/* check if user-supplied timeout has passed */
		if (timeout > 0 && time(NULL) > endtime) {
			char errmsg[GFAL_ERRMSG_LEN];

			if (reqstatp == NULL || (reqstatp->fileStatuses != NULL && reqstatp->fileStatuses->__ptr != NULL))
				for (i = 0; i < reqstatp->fileStatuses->__size; ++i)
					srm_set_xfer_done (srm_endpoint, *reqid, reqstatp->fileStatuses->__ptr[i]->fileId,
							errmsg, GFAL_ERRMSG_LEN, timeout);

			soap_end(&soap);
			soap_done(&soap);

			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: User timeout over",
					gfal_remote_type, srmfunc_status, srm_endpoint);
			errno = ETIMEDOUT;
			return (-1);
		}

	}
	if (reqstatp == NULL || reqstatp->state == NULL || strcasecmp (reqstatp->state, "failed") == 0) {
		if (reqstatp->errorMessage) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: %s",
					gfal_remote_type, srmfunc_status, srm_endpoint, reqstatp->errorMessage);
			sav_errno = errorstring2errno (reqstatp->errorMessage);
		} else  {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
					"[%s][%s][] %s: SRM request failed, but no errorMessage supplied",
					gfal_remote_type, srmfunc_status, srm_endpoint);
			sav_errno = ECOMM;
		}
		soap_end (&soap);
		soap_done (&soap);
		errno = sav_errno;
		return (-1);
	}

	n = reqstatp->fileStatuses->__size;

	if ((*statuses = (struct srm_filestatus *) calloc (n, sizeof(struct srm_filestatus))) == NULL) {
		soap_end (&soap);
		soap_done (&soap);
		errno = ENOMEM;
		return (-1);
	}

	for (i = 0; i < n; i++) {
		memset (*statuses + i, 0, sizeof (struct srm_filestatus));
		if (reqstatp->fileStatuses->__ptr[i]->SURL)
			(*statuses)[i].surl = strdup (reqstatp->fileStatuses->__ptr[i]->SURL);
		(*statuses)[i].fileid = reqstatp->fileStatuses->__ptr[i]->fileId;
		if (!strcasecmp (reqstatp->fileStatuses->__ptr[i]->state, "ready") &&
				reqstatp->fileStatuses->__ptr[i]->TURL) {
			(*statuses)[i].turl = strdup (reqstatp->fileStatuses->__ptr[i]->TURL);
			(*statuses)[i].status = 0;
		} else {
			(*statuses)[i].status = ENOENT;
		}
	}
	soap_end (&soap);
	soap_done (&soap);
	return (n);
}

#if ! defined(linux) || defined(_LARGEFILE64_SOURCE)
srm_getfilemd (int nbfiles, const char **surls, const char *srm_endpoint,
		struct srm_mdfilestatus **mdstatuses, char *errbuf, int errbufsz, int timeout)
{
	int flags;
	struct group *gr;
	struct srm1__getFileMetaDataResponse out;
	struct passwd *pw;
	int ret;
	int sav_errno = 0;
	struct soap soap;
	struct ArrayOfstring surlarray;
	int i, n;
	const char srmfunc[] = "getFileMetaData";

	soap_init (&soap);

#ifdef GFAL_SECURE
	flags = CGSI_OPT_DISABLE_NAME_CHECK;
	soap_register_plugin_arg (&soap, client_cgsi_plugin, &flags);
#endif

	soap.send_timeout = gfal_get_timeout_sendreceive ();
	soap.recv_timeout = gfal_get_timeout_sendreceive ();
	soap.connect_timeout = gfal_get_timeout_connect ();


	/* issue "getFileMetaData" request */

	surlarray.__size = nbfiles;
	surlarray.__ptr = (char **) surls;

	if ((ret = soap_call_srm1__getFileMetaData (&soap, srm_endpoint,
					srmfunc, &surlarray, &out))) {
		if (soap.fault != NULL && soap.fault->faultstring != NULL) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: %s",
					gfal_remote_type, srmfunc, srm_endpoint, soap.fault->faultstring);
			sav_errno = errorstring2errno (soap.fault->faultstring);
		} else if (soap.error == SOAP_EOF) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Connection fails or timeout",
					gfal_remote_type, srmfunc, srm_endpoint);
			sav_errno = ECOMM;
		} else {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Unknown SOAP error (%d)",
					gfal_remote_type, srmfunc, srm_endpoint, soap.error);
			sav_errno = ECOMM;
		}
		soap_end (&soap);
		soap_done (&soap);
		errno = sav_errno;
		return (-1);
	}
	if (out._Result->__size == 0 || out._Result->__ptr[0]->SURL == NULL) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: <empty response>",
				gfal_remote_type, srmfunc, srm_endpoint);
		soap_end (&soap);
		soap_done (&soap);
		errno = ECOMM;
		return (-1);
	}

	n = out._Result->__size;

	if ((*mdstatuses = (struct srm_mdfilestatus *) calloc (n, sizeof (struct srm_mdfilestatus))) == NULL) {
		soap_end (&soap);
		soap_done (&soap);
		errno = ENOMEM;
		return (-1);
	}

	for (i = 0; i < n; ++i) {
		if (out._Result->__ptr[i]->SURL)
			(*mdstatuses)[i].surl = strdup (out._Result->__ptr[i]->SURL);
		(*mdstatuses)[i].stat.st_mode = out._Result->__ptr[i]->permMode;
		if (((*mdstatuses)[i].stat.st_mode & S_IFMT) == 0)
			(*mdstatuses)[i].stat.st_mode |= S_IFREG;
		if (out._Result->__ptr[i]->owner && (pw = getpwnam (out._Result->__ptr[i]->owner)))
			(*mdstatuses)[i].stat.st_uid = pw->pw_uid;
		else
			(*mdstatuses)[i].stat.st_uid = 2;
		if (out._Result->__ptr[i]->group && (gr = getgrnam (out._Result->__ptr[i]->group)))
			(*mdstatuses)[i].stat.st_gid = gr->gr_gid;
		else
			(*mdstatuses)[i].stat.st_gid = 2;
		(*mdstatuses)[i].stat.st_size = out._Result->__ptr[i]->size;
		(*mdstatuses)[i].stat.st_nlink = 1;
	}

	soap_end (&soap);
	soap_done (&soap);
	return (n);
}
#endif

srm_set_xfer_done (const char *srm_endpoint, int reqid, int fileid,	char *errbuf, int errbufsz, int timeout)
{
	return srm_set_xfer_status ("Done", srm_endpoint, reqid, fileid, errbuf, errbufsz, timeout);
}

srm_set_xfer_running (const char *srm_endpoint, int reqid, int fileid, char *errbuf, int errbufsz, int timeout)
{
	return srm_set_xfer_status ("Running", srm_endpoint, reqid, fileid, errbuf, errbufsz, timeout);
}

	static int
srm_set_xfer_status (const char *status, const char *srm_endpoint, int reqid, int fileid,
		char *errbuf, int errbufsz, int timeout)
{
	int flags;
	int sav_errno = 0;
	struct srm1__setFileStatusResponse out;
	struct srmDiskCache__RequestStatus *reqstat;
	int ret;
	struct soap soap;
	const char srmfunc[] = "setFileStatus";

	soap_init (&soap);

#ifdef GFAL_SECURE
	flags = CGSI_OPT_DISABLE_NAME_CHECK;
	soap_register_plugin_arg (&soap, client_cgsi_plugin, &flags);
#endif

	soap.send_timeout = gfal_get_timeout_sendreceive ();
	soap.recv_timeout = gfal_get_timeout_sendreceive ();
	soap.connect_timeout = gfal_get_timeout_connect ();


	if ((ret = soap_call_srm1__setFileStatus (&soap, srm_endpoint,
				srmfunc, reqid, fileid, (char *) status, &out))) {
		if(soap.fault != NULL && soap.fault->faultstring != NULL) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: %s",
					gfal_remote_type, srmfunc, srm_endpoint, soap.fault->faultstring);
			sav_errno = errorstring2errno (soap.fault->faultstring);
		} else if (soap.error == SOAP_EOF) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Connection fails or timeout",
					gfal_remote_type, srmfunc, srm_endpoint);
			sav_errno = ECOMM;
		} else {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: Unknown SOAP error (%d)",
					gfal_remote_type, srmfunc, srm_endpoint, soap.error);
			sav_errno = ECOMM;
		}
		soap_end (&soap);
		soap_done (&soap);
		errno = sav_errno;
		return (-1);
	}

	if ((reqstat = out._Result) == NULL || reqstat->state == NULL || strcasecmp (reqstat->state, "failed") == 0) {
		if (reqstat->errorMessage) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[%s][%s][] %s: %s",
					gfal_remote_type, srmfunc, srm_endpoint, reqstat->errorMessage);
			sav_errno = errorstring2errno (reqstat->errorMessage);
		} else  {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
					"[%s][%s][] %s: SRM request failed, but no errorMessage supplied",
					gfal_remote_type, srmfunc, srm_endpoint);
			sav_errno = ECOMM;
		}
		soap_end (&soap);
		soap_done (&soap);
		errno = sav_errno;
		return (-1);
	}

	soap_end (&soap);
	soap_done (&soap);
	return (0);
}

	static int
errorstring2errno (const char *errstr)
{
	if (!errstr)
		return (ECOMM);
	else if (strstr (errstr, "ile exists"))
		return (EEXIST);
	else if (strstr (errstr, "does not exist") ||
			strstr (errstr, "o such file or directory") ||
			strstr (errstr, "could not get storage info by path"))
		return (ENOENT);
	else if (strstr (errstr, "ermission denied"))
		return (EACCES);
	else if (strstr (errstr, "nvalid arg") ||
			strstr (errstr, "nknown"))
		return (EINVAL);
	else if (strstr (errstr, "rotocol"))
		return (EPROTONOSUPPORT);
	else if (strstr (errstr, "o space left on device"))
		return (ENOSPC);
	else
		return (ECOMM);
}
